from scipy import sparse
import matplotlib.pyplot as plt
from math import gcd
import itertools
import numpy as np
from shapely.geometry import Point, LineString, Polygon
import cvxpy as cvx
from enum import IntEnum

# define boundary conditions
class BD(IntEnum):
    Fix = 0
    Internal = 1
    Free = 2
    Symmetry = 3
    Load = 4
    Rigid_load = 5
    Lat_load = 6

# solve DLO problem
def DLO(vt, edgebd, mat):
    poly = Polygon(vt)
    Nd = createNodes(poly)
    PML = createDiscontinuities(poly, Nd)
    bdL = setCnBoundaryCondition(Nd, PML, vt, edgebd)
    print('Nodes: %d Discontinuities: %d' % (len(Nd), len(PML)))
    PML[(PML[:, 2] <= 1.42) | (bdL != BD.Internal), 3] = True
    # Start the adaptive solution loop
    for itr in range(1, 100):
        activeCn = np.where(PML[:, 3] == True)[0]
        Cn = PML[activeCn]
        bd = bdL[activeCn]
        factor, d, p, t = solveLP(Nd, Cn, bd, mat)
        print("Itr: %d, Load factor: %f, Active disc.: %d" % (itr, factor, len(Cn)))
        if stopViolation(factor, t, Nd, PML, bdL, mat):
            break
    plot(vt, d, Cn, Nd, bd)
    plotDispField(vt, d, Cn, Nd, bd, spacing=1 ,maxSize=0.8)
    plotBoundaries(vt, edgebd, symbolSize=1/20)
    plt.show()
    return factor

# Solve the Linear Programming problem
def solveLP(Nd, Cn, bd, mat):
    m, n = len(Cn), len(Nd)
    B = calcB(Nd, Cn)
    N = calcN(bd, mat['phi'])
    g = calcg(Cn, mat['cohesive'])
    fL, fD = loadEffect(Nd, Cn, bd, mat['weight'])
    activeN, activeG = boundaryConditions(bd)
    d = cvx.Variable(2 * m)
    p = cvx.Variable(2 * m)
    energy = - fD.transpose() @ d + (g * activeG).transpose() @ p
    obj = cvx.Minimize(energy)
    cons = []
    compatibility = (B @ d == 0)
    cons.append(compatibility)  # Compatibility constraints
    cons.append(fL @ d == 1)  # Impose unitary live loads work
    cons.append(p >= 0)  # Impose positive plastic multipliers
    cons.append(cvx.multiply(N @ p - d, activeN) == 0)  # Flow rule constraints
    linkN, linkS = processCnLinks(bd)  # Rigid load constraints
    if len(linkN):
        nL = d[linkN * 2 + 1]
        cons.append(nL[1:] == nL[:-1])
    if len(linkS):
        sL = d[linkS * 2]
        cons.append(sL[1:] == sL[:-1])
    prob = cvx.Problem(obj, cons)
    factor = prob.solve(solver='ECOS')
    d = np.array(d.value).flatten()
    p = np.array(p.value).flatten()
    t = -np.array(compatibility.dual_value)
    return factor, d, p, t

# Check dual violation
def stopViolation(factor, t, Nd, PML, boundaries, material):
    inactiveCnlst = np.where(PML[:, 3] == False)[0]
    Cn = PML[inactiveCnlst]
    bd = boundaries[inactiveCnlst]
    m = len(Cn)
    B = calcB(Nd, Cn).tocsc()
    N = calcN(bd, material['phi']).tocsc()
    g = calcg(Cn, material['cohesive'] + 1e-16)
    fL, fD = loadEffect(Nd, Cn, bd, material['weight'])
    q = B.transpose().dot(t) + factor * fL + fD
    y = N.transpose().dot(q) / g
    vio1 = y[np.arange(0, m * 2, 2)]
    vio2 = y[np.arange(1, m * 2, 2)]
    vio = np.vstack((vio1, vio2))
    vio = np.amax(vio, axis=0)
    vioCn = np.where(vio > 1.0001)[0]
    vio = vio[vioCn]
    vioSort = np.flipud(np.argsort(vio))
    mv = len(vioSort)
    mp = len(Cn)
    kappa1, kappa2 = 0.05, 0.05
    if mv >= kappa2 * mp:
        num = kappa1 * mv
    elif kappa1 * kappa2 * mp >= mv:
        num = mv
    else:
        num = kappa1 * kappa2 * mp
    num = int(np.ceil(num))
    for i in range(num):
        inactiveCn = vioCn[vioSort[i]]
        Cid = inactiveCnlst[inactiveCn]
        PML[Cid][3] = True
    return num == 0

# Calculate the g vector
def calcg(Cn, c):
    l = Cn[:, 2]
    g = np.vstack((l*c, l*c)).flatten(order='F')
    return g

# Calculate compatibility matrix B
def calcB(Nd, Cn):
    m, n = len(Cn), len(Nd)
    n1, n2 = Cn[:, 0].astype(int), Cn[:, 1].astype(int)
    l, X, Y = Cn[:, 2], Nd[n2, 0] - Nd[n1, 0], Nd[n2, 1] - Nd[n1, 1]
    alpha, beta = X / l, Y / l
    values = [[alpha, -beta],
              [beta, alpha],
              [-alpha, beta],
              [-beta, -alpha]]
    rows = [n1 * 2,
            n1 * 2 + 1,
            n2 * 2,
            n2 * 2 + 1]
    columns = [np.arange(0, 2 * m, 2), np.arange(1, 2 * m, 2)]
    return toSparseMatrix(values, rows, columns,
                          shape=(n * 2, m * 2))

# Calculate flow rule matrix N
def calcN(boundaries, phi):
    m = len(boundaries)
    mu = np.tan(phi / 180 * np.pi) * np.ones(m)
    one = np.ones(m)
    values = [[one, -one],
              [mu, mu]]
    columns = [np.arange(0, 2 * m, 2), np.arange(1, 2 * m, 2)]
    rows = [np.arange(0, 2 * m, 2), np.arange(1, 2 * m, 2)]
    return toSparseMatrix(values, rows, columns,
                          shape=(m * 2, m * 2))

# Calculate load effects
def loadEffect(Nd, Cn, bd, weight):
    n1, n2 = Cn[:, 0].astype(int), Cn[:, 1].astype(int)
    m, l = len(Cn), Cn[:, 2]
    ym = 0.5 * (Nd[n2, 1] + Nd[n1, 1])
    loadedCn = np.where( (bd == BD.Load) | (bd==BD.Rigid_load) | (bd==BD.Lat_load))[0]
    fL, fD = np.zeros((m, 2)), np.zeros((m, 2))
    fL[loadedCn, 1] = l[loadedCn]
    fD[:, 1] = l*weight*ym
    return fL.flatten(), fD.flatten()

# check flow rules
def hasFlowrule(boundary):
    if boundary in [BD.Internal, BD.Fix]:
        s, n = 1, 1
    elif boundary in [BD.Symmetry]:
        s, n = 0, 1
    elif boundary in [BD.Lat_load]:
        s, n = 1, 0
    else:
        s, n = 0, 0
    return [s, n]

# check internal energy dissipation
def hasEnergy(bd):
    if bd in [BD.Internal, BD.Fix, BD.Symmetry, BD.Lat_load]:
        return [1, 1]
    else:
        return [0, 0]

# process boundary conditions
def boundaryConditions(boundaries):
    activeN = np.array([hasFlowrule(boundary)
                        for boundary in boundaries])
    activeG = np.array([hasEnergy(boundary)
                        for boundary in boundaries])
    return activeN.flatten(), activeG.flatten()

# identify linked discontinuities
def processCnLinks(bd):
    LinkN = np.where((bd == BD.Rigid_load) | (bd == BD.Lat_load))[0]
    if len(LinkN) and BD.Rigid_load in bd[LinkN]:
        LinkS = LinkN #rigid load
    else:
        LinkS = [] #laterally constrained rigid load
    return LinkN, LinkS

# create sparse matrix
def toSparseMatrix(values, rows, columns, shape):
    v, r, c = np.array(values), \
              np.tile(rows, (1, len(columns))), \
              np.tile(columns, (len(rows), 1))
    return sparse.coo_matrix((v.flatten(), (r.flatten(), c.flatten())), shape=shape)

# generate nodes
def createNodes(poly):
    xv, yv = np.meshgrid(np.arange(poly.bounds[0], poly.bounds[2] + 1).astype(int),
                         np.arange(poly.bounds[1], poly.bounds[3] + 1).astype(int))
    pts = [Point(xv.flat[i], yv.flat[i]) for i in range(xv.size)]
    Nd = np.array([[pt.x, pt.y] for pt in pts if poly.intersects(pt)])
    return Nd

# generate discontinuities
def createDiscontinuities(poly, Nd):
    PML = []
    convex = True if poly.convex_hull.area == poly.area else False
    for i, j in itertools.combinations(range(len(Nd)), 2):
        dx, dy = abs(Nd[i][0] - Nd[j][0]), abs(Nd[i][1] - Nd[j][1])
        if gcd(int(dx), int(dy)) == 1:
            seg = [] if convex else LineString([Nd[i], Nd[j]])
            if convex or poly.contains(seg) or poly.boundary.contains(seg):
                PML.append([i, j, np.sqrt(dx ** 2 + dy ** 2), False])
    return np.array(PML)

# set boundary conditions to discontinuities
def setCnBoundaryCondition(Nd, Cn, vt, edgebd):
    n, m, vtnum = len(Nd), len(Cn), len(vt)
    n1, n2 = Cn[:,0].astype(int), Cn[:,1].astype(int)
    Bd_Nd = np.zeros(n).astype('uint64')
    #identify all boundary nodes
    for i in range(len(vt)):
        v1, v2 = vt[i], vt[(i+1)%vtnum]
        edgeLine = LineString([v1, v2])
        for j in range(n):
            pt = Point(Nd[j, 0], Nd[j, 1])
            if edgeLine.intersects(pt):
                Bd_Nd[j] += 2 ** i
    #use binary and to identify discontinuities on edges
    Bd_n1 = Bd_Nd[n1].astype('uint64')
    Bd_n2 = Bd_Nd[n2].astype('uint64')
    Bd_Cn = np.bitwise_and(Bd_n1, Bd_n2)
    edgeCn = np.where(Bd_Cn)[0]
    edgeId = np.log2(Bd_Cn[edgeCn]).astype(int)
    bd = np.ones(m) * BD.Internal
    bd[edgeCn] = np.array(edgebd)[edgeId]
    return bd

# filter discontinuities with no relative displacements
def filter(Cn, d, bd, ratio=1e-4):
    boundaryID = [bi != BD.Internal for bi in bd]
    m = len(Cn)
    s = d[range(0, m * 2, 2)]
    n = d[range(1, m * 2, 2)]
    maxD = max(max(abs(s)), max(abs(n)))
    threshold = maxD * ratio
    validCn = np.zeros(m)
    validCn[abs(s) > threshold] = 1
    validCn[abs(n) > threshold] = 1
    validCn[np.where(boundaryID)[0]] = 1
    Cnlst = np.where(validCn)[0]
    return Cnlst

# plot failure mechanism
def plot(vt, d, Cn, Nd, bd):
    plt.axis('off')
    plt.axis('equal')
    plt.draw()
    ax = plt.gca()
    ax.add_patch(plt.Polygon(vt, facecolor=[0.9, 0.9, 0.9], edgecolor='k', alpha=1))
    Cnlst = filter(Cn, d, bd)
    maxD = np.max(np.abs(d))
    for i in Cnlst:
        internalWork = abs(np.dot(d[[i*2,i*2+1]], hasFlowrule(bd[i])))
        if internalWork>1e-6:
            width = 6 * (abs(d[i * 2]) + abs(d[i * 2 + 1])) / 2 / maxD
            n1, n2 = Cn[i][0].astype(int), Cn[i][1].astype(int)
            plt.plot((Nd[n1][0], Nd[n2][0]),
                     (Nd[int(Cn[i][0])][1], Nd[int(Cn[i][1])][1]),
                     'r', linewidth=width+1, zorder=10)

# plat displacement field
def plotDispField(vt, d, Cn, Nd, bd, spacing = 1, maxSize = 1):
    xmin, ymin = np.min(vt, axis=0)
    xmax, ymax = np.max(vt, axis=0)
    xdiv = int((xmax-xmin)/spacing)
    ydiv = int((ymax-ymin)/spacing)
    x, y = np.linspace(xmin, xmax, xdiv), np.linspace(ymin, ymax, ydiv)
    xgrid, ygrid = np.meshgrid(x, y)
    u, v = np.zeros(xgrid.shape), np.zeros(xgrid.shape)
    m = len(Cn)
    s, n = d[range(0, m * 2, 2)], d[range(1, m * 2, 2)]
    ids = filter(Cn, d, bd)
    Cn, s, n = Cn[ids], s[ids], n[ids]
    maxD = np.max(np.abs(d))
    s /= maxD
    n /= maxD
    n1, n2 = Cn[:, 0].astype(int), Cn[:, 1].astype(int)
    x1, y1, x2, y2 = Nd[n1, 0], Nd[n1, 1], Nd[n2, 0], Nd[n2, 1]
    tol = 1e-3
    for i in range(len(y)):
        yi = y[i]
        # find x coordinates of all intersection points
        crossX = (yi - y1) * (x2 - x1) / (y2 - y1 + tol) + x1
        crossID = np.where(abs(y1 - yi + tol) + abs(y2 - yi + tol) < abs(y1 - y2) + tol)[0]
        crossX = crossX[crossID]
        for j in range(len(crossID)):
            cid = crossID[j]
            lst = np.where(x > crossX[j] - tol)[0]
            l, X, Y = Cn[cid, 2], x2[cid] - x1[cid], y2[cid] - y1[cid]
            alpha, beta = X / l, Y / l
            u[i, lst] -= alpha * s[cid] - beta * n[cid]
            v[i, lst] -= beta * s[cid] + alpha * n[cid]
    maxVecLength = np.max(np.sqrt(u ** 2 + v ** 2)) / maxSize
    u /= maxVecLength
    v /= maxVecLength
    ID = np.where( (abs(u)>tol) | (abs(v)>tol))
    plt.quiver(xgrid[ID], ygrid[ID], u[ID], v[ID], color='g',
               units = 'xy', scale_units='xy', scale = 1,
               minlength=0, width=0.08, zorder=5)

# plot boundary conditions
def plotBoundaries(vt, boundaries, symbolSize = 1/20):
    size = np.min(np.max(vt, axis=0)) * symbolSize
    for i in range(len(boundaries)):
        type, p1, p2 = boundaries[i], vt[i], vt[(i + 1) % len(vt)]
        plotEdgeBoundary(p1, p2, type, size, linewidth=1)

# helper function to plot boundaries
def plotEdgeBoundary(v1, v2, type, symbolsize=1, linewidth=1):
    ax = plt.gca()
    vangle = np.arctan2(v2[1] - v1[1], v2[0] - v1[0])
    vlength = np.linalg.norm([v2[1] - v1[1], v2[0] - v1[0]])
    a1 = vangle - np.pi / 4
    a2 = vangle - np.pi * 3 / 4
    num = np.round(vlength / symbolsize + 1).astype(int)
    if type == BD.Fix:
        pos = np.linspace(v1, v2, num)
        for i in range(0, num - 1):
            pend = pos[i] + np.array([np.cos(a1), np.sin(a1)]) * np.sqrt(2) * symbolsize
            ax.plot([pos[i, 0], pend[0]], [pos[i, 1], pend[1]], 'k', linewidth=linewidth)
        for i in np.arange(1, num):
            pend = pos[i] + np.array([np.cos(a2), np.sin(a2)]) * np.sqrt(2) * symbolsize
            ax.plot([pos[i, 0], pend[0]], [pos[i, 1], pend[1]], 'k', linewidth=linewidth)

    if type == BD.Symmetry:
        offset = np.array([np.cos(vangle - np.pi / 2), np.sin(vangle - np.pi / 2)]) * symbolsize
        v3, v4 = v2 + offset, v1 + offset
        ax.plot([v3[0], v4[0]], [v3[1], v4[1]], 'k',
                linestyle=(0, (symbolsize*20, symbolsize*10, 1, symbolsize*10)),
                linewidth=linewidth)

    if type in [BD.Load, BD.Rigid_load, BD.Lat_load]:
        offset1 = np.array([np.cos(vangle - np.pi / 2), np.sin(vangle - np.pi / 2)]) * np.sqrt(2) * symbolsize * 2
        if type is not BD.Load:
            offset2 = offset1 / 2
            vt = [v1, v2, v2+offset2, v1+offset2]
            ax.add_patch(plt.Polygon(vt, facecolor=[0.5, 0.5, 0.5], edgecolor='k', alpha=1))
        else:
            offset2 = 0
        pos = np.linspace(v1, v2, int(num/2))
        for p in pos:
            tail = p + offset1 + offset2
            ax.arrow(tail[0], tail[1], -offset1[0], -offset1[1],
                      head_width=symbolsize, length_includes_head=True, color='k')
