from scipy import sparse
import matplotlib.pyplot as plt
from math import gcd
import itertools
import numpy as np
from shapely.geometry import Point, LineString, Polygon
import cvxpy as cvx
from enum import IntEnum

# define boundary conditions
class BD(IntEnum):
    Fix = 0
    Internal = 1
    Free = 2
    Symmetry = 3
    Load = 4

# solve DLO problem
def DLO(vt, edgebd, mat):
    poly = Polygon(vt)
    Nd = createNodes(poly)
    Cn = createDiscontinuities(poly, Nd)
    bd = setCnBoundaryCondition(Nd, Cn, vt, edgebd)
    factor, d, p = solveLP(Nd, Cn, bd, mat)
    plot(vt, d, Cn, Nd, bd)
    return factor

# Solve the Linear Programming problem
def solveLP(Nd, Cn, bd, mat):
    m, n = len(Cn), len(Nd)
    B = calcB(Nd, Cn)
    N = calcN(bd)
    g = calcg(Cn, mat['cohesive'])
    fL = unitLoad(Cn, bd)
    activeN, activeG = boundaryConditions(bd)
    d = cvx.Variable(2 * m)
    p = cvx.Variable(2 * m)
    energy = (g * activeG).transpose() @ p
    obj = cvx.Minimize(energy)
    cons = []
    cons.append(B @ d == 0)  # Compatibility constraints
    cons.append(fL @ d == 1)  # Impose unitary live loads work
    cons.append(p >= 0)  # Impose positive plastic multipliers
    cons.append(cvx.multiply(N @ p - d, activeN) == 0)  # Flow rule constraints
    prob = cvx.Problem(obj, cons)
    factor = prob.solve(solver='ECOS')
    d = np.array(d.value).flatten()
    p = np.array(p.value).flatten()
    return factor, d, p

# Calculate the g vector
def calcg(Cn, c):
    l = Cn[:, 2]
    g = np.array([[li * c, li * c] for li in l])
    return g.flatten()

# Calculate compatibility matrix B
def calcB(Nd, Cn):
    m, n = len(Cn), len(Nd)
    n1, n2 = Cn[:, 0].astype(int), Cn[:, 1].astype(int)
    l, X, Y = Cn[:, 2], Nd[n2, 0] - Nd[n1, 0], Nd[n2, 1] - Nd[n1, 1]
    cos_x, cos_y = X / l, Y / l
    values = [[cos_x, -cos_y],
              [cos_y, cos_x],
              [-cos_x, cos_y],
              [-cos_y, -cos_x]]
    rows = [n1 * 2,
            n1 * 2 + 1,
            n2 * 2,
            n2 * 2 + 1]
    columns = [np.arange(0, 2 * m, 2), np.arange(1, 2 * m, 2)]
    return toSparseMatrix(values, rows, columns, shape=(n * 2, m * 2))

# Calculate flow rule matrix N
def calcN(boundaries):
    m = len(boundaries)
    one = np.ones(m)
    zero = np.zeros(m)
    values = [[one, -one],
              [zero, zero]]
    columns = [np.arange(0, 2 * m, 2), np.arange(1, 2 * m, 2)]
    rows = [np.arange(0, 2 * m, 2), np.arange(1, 2 * m, 2)]
    return toSparseMatrix(values, rows, columns, shape=(m * 2, m * 2))

# Calculate load effects
def unitLoad(Cn, bd):
    m, l = len(bd), Cn[:,2]
    loadedCn = np.where(bd == BD.Load)[0]
    fL = np.zeros((m, 2))
    fL[loadedCn, 1] = l[loadedCn]
    return fL.flatten()

# check flow rules
def hasFlowrule(boundary):
    if boundary in [BD.Internal, BD.Fix]:
        s, n = 1, 1
    elif boundary in [BD.Symmetry]:
        s, n = 0, 1
    else:
        s, n = 0, 0
    return [s, n]

# check internal energy dissipation
def hasEnergy(bd):
    if bd in [BD.Internal, BD.Fix, BD.Symmetry]:
        return [1, 1]
    else:
        return [0, 0] #no energy dissipation

# process boundary conditions
def boundaryConditions(boundaries):
    activeN = np.array([hasFlowrule(boundary) for boundary in boundaries])
    activeG = np.array([hasEnergy(boundary) for boundary in boundaries])
    return activeN.flatten(), activeG.flatten()

# create sparse matrix
def toSparseMatrix(values, rows, columns, shape):
    v, r, c = np.array(values), \
              np.tile(rows, (1, len(columns))), \
              np.tile(columns, (len(rows), 1))
    return sparse.coo_matrix((v.flatten(), (r.flatten(), c.flatten())), shape=shape)

# generate nodes
def createNodes(poly):
    xv, yv = np.meshgrid(np.arange(poly.bounds[0], poly.bounds[2] + 1).astype(int),
                         np.arange(poly.bounds[1], poly.bounds[3] + 1).astype(int))
    pts = [Point(xv.flat[i], yv.flat[i]) for i in range(xv.size)]
    Nd = np.array([[pt.x, pt.y] for pt in pts if poly.intersects(pt)])
    return Nd

# generate discontinuities
def createDiscontinuities(poly, Nd):
    Cn = []
    convex = True if poly.convex_hull.area == poly.area else False
    for i, j in itertools.combinations(range(len(Nd)), 2):
        dx, dy = abs(Nd[i][0] - Nd[j][0]), abs(Nd[i][1] - Nd[j][1])
        if gcd(int(dx), int(dy)) == 1:
            seg = [] if convex else LineString([Nd[i], Nd[j]])
            if convex or poly.contains(seg) or poly.boundary.contains(seg):
                Cn.append([i, j, np.sqrt(dx ** 2 + dy ** 2)])
    return np.array(Cn)

# set boundary conditions to discontinuities
def setCnBoundaryCondition(Nd, Cn, vt, edgebd):
    m = len(Cn)
    bd = np.ones(m) * BD.Internal
    for i in range(m):
        for j in range(len(vt)):
            type, p1, p2 = edgebd[j], vt[j], vt[(j + 1) % len(vt)]
            pointA = Nd[int(Cn[i, 0])]
            pointB = Nd[int(Cn[i, 1])]
            if LineString([p1, p2]).contains(LineString([(pointA), (pointB)])):
                bd[i] = type
                break
    return np.array(bd)

# plot failure mechanism
def plot(vt, d, Cn, Nd, bd):
    plt.axis('off')
    plt.axis('equal')
    plt.draw()
    ax = plt.gca()
    ax.add_patch(plt.Polygon(vt, facecolor=[0.9, 0.9, 0.9], edgecolor='k', alpha=1))
    maxD = np.max(np.abs(d))
    threshold = maxD * 1e-4
    for i in range(len(Cn)):
        if abs(d[i * 2]) > maxD * threshold or abs(d[i * 2 + 1]) > maxD * threshold:
            if bd[i] == BD.Internal:
                width = 6 * (abs(d[i * 2]) + abs(d[i * 2 + 1])) / 2 / maxD
                n1, n2 = Cn[i][0].astype(int), Cn[i][1].astype(int)
                plt.plot((Nd[n1][0], Nd[n2][0]),
                         (Nd[int(Cn[i][0])][1], Nd[int(Cn[i][1])][1]), 'r', linewidth=width)
    plt.show()
